@@ -1,134 +1,72 @@
 
-// balance - a balancing tcp proxy
-// $Revision: 2.33 $
-//
-// Copyright (c) 2000,2001 by Thomas Obermair (obermair@acm.org) 
-// and Inlab Software GmbH (info@inlab.de), Gruenwald, Germany. 
-// All rights reserved.
-//
-// Thanks to Bernhard Niederhammer for the initial idea and heavy
-// testing on *big* machines ... 
-//
-// For license terms, see the file COPYING in this directory.
-// 
-// This program is dedicated to Richard Stevens... 
+/*
+ * balance - a balancing tcp proxy
+ * "Professional" Release Branch
+ * $Revision: 3.15 $
+ *
+ * Copyright (c) 2000-2003,2004 by Thomas Obermair (obermair@acm.org)
+ * and Inlab Software GmbH (info@inlab.de), Gruenwald, Germany.
+ * All rights reserved.
+ *
+ * Thanks to Bernhard Niederhammer for the initial idea and heavy
+ * testing on *big* machines ...
+ *
+ * For license terms, see the file COPYING in this directory.
+ *
+ * This program is dedicated to Richard Stevens...
+ *
+ *  3.15
+ *    first -B support
+ *  3.14
+ *    -Wall cleanup
+ *  3.12
+ *    alarm(0) added, thanks to Jon Christensen 
+ *  3.11
+ *    Bugfix
+ *  3.10
+ *    Bugfix for RedHat 7.2
+ *  3.9
+ *    Moved rendezvous file to /var/run and cleaned main(), thanks to Kayne Naughton
+ *  3.8
+ *    move to sigaction(), thanks to Kayne Naughton
+ *  3.5
+ *    Select-Timeout, thanks to Jeff Buhlmann
+ *  3.2
+ *    Hash groups and some other improvements
+ *  2.24:
+ *    'channel 2 overload' problem fixed, thanks to Ed "KuroiNeko"
+ *  2.26:
+ *    'endless loop error' fixed, thanks to Anthony Baxter
+ *  2.27:
+ *    strcmp on NULL removed, thanks to Jay. D. Allen
+ *  2.28:
+ *    bsent and breceived now unsigned to avoid negative values,
+ *    thanks to Anthony Baxter
+ *  2.29:
+ *    error in setaddress() fixed, thanks to Dirk Datzert
+ *  2.30:
+ *    fixing #includes for *bsd compability
+ *  2.31:
+ *  2.32:
+ *    redefied SIGCHLD handling to be compatible with FreeBSD 4.3,
+ *    BSD/OS 4.2 and BSD/OS 4.0.1
+ *  2.33
+ *    finally included SO_REUSEADDR
+ *
+ */
 
-// Changes:
-//
-// 2.24:
-//   'channel 2 overload' problem fixed, thanks to Ed "KuroiNeko"
-// 2.26:
-//   'endless loop error' fixed, thanks to Anthony Baxter
-// 2.27:
-//   strcmp on NULL removed, thanks to Jay. D. Allen
-// 2.28: 
-//   bsent and breceived now unsigned to avoid negative values,
-//   thanks to Anthony Baxter
-// 2.29:
-//   error in setaddress() fixed, thanks to Dirk Datzert
-// 2.30:
-//   fixing #includes for *bsd compability
-// 2.31: 
-// 2.32:
-//   redefied SIGCHLD handling to be compatible with FreeBSD 4.3,
-//   BSD/OS 4.2 and BSD/OS 4.0.1 
-// 2.33
-//   finally included SO_REUSEADDR
-//
+#include <balance.h>
 
-#include <stdio.h>
-#include <strings.h>
-#include <ctype.h>
-#include <errno.h>
-#include <netdb.h>
-#include <signal.h>
-#include <stdlib.h>
+const char* balance_rcsid="$Id: balance.c,v 3.15 2004/05/27 12:18:23 tommy Exp $";
+static char* revision="$Revision: 3.15 $";
 
-#ifdef __FreeBSD__
-#define MYBSD 1
-#endif 
-
-#ifdef bsdi
-#define MYBSD 1
-#endif
-
-#ifdef BSD
-#define MYBSD 1
-#endif
-
-#ifdef MYBSD
-#include <sys/wait.h>
-#else
-#include <sys/resource.h>
-#endif
-
-#include <sys/socket.h>
-#include <sys/types.h>
-#include <sys/wait.h>
-#include <sys/stat.h>
-#include <fcntl.h>
-#include <unistd.h>
-#include <string.h>
-#include <sys/shm.h>
-
-#include <netinet/in.h>
-#include <arpa/inet.h>
-
-#define MAXTXSIZE 	(32*1024)
-#define FILENAMELEN 	1024
-#define SHMDIR 		"/tmp/"
-
-#define MAXCHANNELS 	16	// max channels in group 
-#define MAXGROUPS   	16	// max groups
-#define MAXINPUTLINE 	128	// max line in input mode 
-#define DEFAULTTIMEOUT  5	// default-timeout for non reachable hosts
-
-static char* rcsid="$Id: balance.c,v 2.33 2001/09/15 22:00:54 tommy Exp $";
-static char* revision="$Revision: 2.33 $";
+static int release;
+static int subrelease;
 
 static char rendezvousfile[FILENAMELEN];
 static int  rendezvousfd;	
 
-typedef struct {
-  int status;
-  int port;
-  struct in_addr ipaddr;
-  int c;			// current # of connections
-  int maxc;			// max # of connections, 0 = no limit
-  unsigned int bsent;		// bytes sent
-  unsigned int breceived;	// bytes received
-} CHANNEL;
-
-typedef struct {
-  int nchannels;		// number of channels in group
-  int current;			// current channel in group
-  CHANNEL channels[MAXCHANNELS];
-} GROUP;
-
-typedef struct {
-  GROUP groups[MAXGROUPS];
-} COMMON;
-
-// Macros to access various elements of struct GROUP and struct CHANNEL 
-// within COMMON array
-//
-// a       pointer to variable of type COMMON
-// g       group index
-// i       channel index
-
-#define cmn_group(a,g)       	 ((a)->groups[(g)])
-#define grp_nchannels(a,g) 	 (cmn_group((a),(g)).nchannels)
-#define grp_current(a,g)    	 (cmn_group((a),(g)).current)
-#define grp_channel(a,g,i)  	 (cmn_group((a),(g)).channels[(i)])
-#define chn_status(a,g,i)   	 (grp_channel((a),(g),(i)).status)
-#define chn_port(a,g,i)		 (grp_channel((a),(g),(i)).port)
-#define chn_ipaddr(a,g,i)	 (grp_channel((a),(g),(i)).ipaddr)
-#define chn_c(a,g,i)		 (grp_channel((a),(g),(i)).c)
-#define chn_maxc(a,g,i)		 (grp_channel((a),(g),(i)).maxc)
-#define chn_bsent(a,g,i)	 (grp_channel((a),(g),(i)).bsent)
-#define chn_breceived(a,g,i)	 (grp_channel((a),(g),(i)).breceived)
-
-static err_dump(char *text) {
+int static err_dump(char *text) {
   fprintf(stderr,"balance: %s\n",text);
   fflush(stderr);
   exit(1);
@@ -143,9 +81,12 @@ static int interactive;
 
 static int connect_timeout;
 
+static char* bindhost=NULL;
+static char* outbindhost=NULL;
+
 // locking ...
 
-a_readlock(off_t start, off_t len) {
+int a_readlock(off_t start, off_t len) {
   int rc;
   struct flock fdata;
   fdata.l_type=F_RDLCK;
@@ -166,13 +107,13 @@ repeat:
   return(rc);
 }
 
-b_readlock() { a_readlock(0,0); }
+void b_readlock() { a_readlock(0,0); }
 
-c_readlock(int group, int channel) { 
+void c_readlock(int group, int channel) { 
   a_readlock(((char*)&(grp_channel(common,group,channel)))-(char*)common,sizeof(CHANNEL)); 
 }
 
-a_writelock(off_t start, off_t len) {
+int a_writelock(off_t start, off_t len) {
   int rc;
   struct flock fdata;
   fdata.l_type=F_WRLCK;
@@ -193,13 +134,13 @@ repeat:
   return(rc);
 }
 
-b_writelock() { a_writelock(0,0); }
+void b_writelock() { a_writelock(0,0); }
 
-c_writelock(int group, int channel) { 
+void c_writelock(int group, int channel) { 
   a_writelock(((char*) &(grp_channel(common,group,channel)))-(char*)common,sizeof(CHANNEL)); 
 }
 
-a_unlock(off_t start, off_t len) {
+int a_unlock(off_t start, off_t len) {
   int rc;
   struct flock fdata;
   fdata.l_type=F_UNLCK;
@@ -220,9 +161,9 @@ repeat:
   return(rc);
 }
 
-b_unlock() { a_unlock(0,0); }
+void b_unlock() { a_unlock(0,0); }
 
-c_unlock(int group, int channel) { 
+void c_unlock(int group, int channel) { 
   a_unlock(((char*) &(grp_channel(common,group,channel)))-(char*)common,sizeof(CHANNEL)); 
 }
 
@@ -232,7 +173,6 @@ void* shm_malloc(char* file, int size) {
   key_t key;
   int shmid;
   char *data;
-  int mode;
 
   if ((key = ftok(file, 'x')) == -1) {
     perror("ftok");
@@ -294,7 +234,7 @@ int getport(char* port) {
   }
 }
 
-int setipaddress(struct in_addr *ipaddr, char* string) {
+void setipaddress(struct in_addr *ipaddr, char* string) {
   struct hostent *hent;
   hent = gethostbyname(string);
   if(hent == NULL) {
@@ -434,8 +374,7 @@ int writen(int fd, char* ptr, int nbytes) {
   return(nbytes - nleft);
 }
 
-hin(int von, int nach, int groupindex, int channelindex) {
-  char c;
+int hin(int von, int nach, int groupindex, int channelindex) {
   int rc;
 
   char buffer[MAXTXSIZE];
@@ -460,8 +399,7 @@ hin(int von, int nach, int groupindex, int channelindex) {
   return(0);
 }
 
-zurueck(int von, int nach, int groupindex, int channelindex) {
-  char c;
+int zurueck(int von, int nach, int groupindex, int channelindex) {
   int rc;
 
   char buffer[MAXTXSIZE];
@@ -489,7 +427,13 @@ zurueck(int von, int nach, int groupindex, int channelindex) {
 // the connection is really established, let's transfer the data
 // as efficient as possible :-) 
 
-stream2(int clientfd, int serverfd, int groupindex, int channelindex) {
+static struct timeval sel_tmout = {
+  0, 	// seconds
+  0	// microseconds
+};
+static struct timeval save_tmout = { 0, 0 };
+
+void stream2(int clientfd, int serverfd, int groupindex, int channelindex) {
   fd_set readfds;
   int    fdset_width;
   int    sr;
@@ -501,9 +445,17 @@ stream2(int clientfd, int serverfd, int groupindex, int channelindex) {
     FD_ZERO(&readfds);
     FD_SET(clientfd,&readfds);
     FD_SET(serverfd,&readfds);
+    sel_tmout = save_tmout;
 
     for(;;) {
-      sr=select(fdset_width, &readfds, NULL, NULL, NULL); 
+      if(sel_tmout.tv_sec || sel_tmout.tv_usec) 
+        sr=select(fdset_width, &readfds, NULL, NULL, &sel_tmout); 
+      else
+        sr=select(fdset_width, &readfds, NULL, NULL, NULL); 
+      if((save_tmout.tv_sec || save_tmout.tv_usec) && !sr) {
+        fprintf(stderr,"timed out after %d seconds\n",(int) save_tmout.tv_sec);
+        exit(1); 
+      }
       if (sr < 0 && errno != EINTR) {
         err_dump("select error");
       }
@@ -527,23 +479,25 @@ stream2(int clientfd, int serverfd, int groupindex, int channelindex) {
   exit(1);
 }
 
-void alarm_handler() {
+void alrm_handler(int signo) {
 }
 
-void chld_handler() {
+void usr1_handler(int signo) {
+}
+
+void chld_handler(int signo) {
   int status;
   while(waitpid(-1, &status, WNOHANG) > 0);
-  (void) signal(SIGCHLD, chld_handler);
 }
 
 // a channel in a group is selected and we try to establish a connection 
 
-void *stream(int arg, int groupindex, int index) {
+void *stream(int arg, int groupindex, int index, char* client_address, int client_address_size) {
   int			  startindex;
   int                     sockfd;
   int			  clientfd;
+  struct sigaction        alrm_action;
   struct sockaddr_in      serv_addr;
-  char id[64];
 
   startindex=index;	  // lets keep where we start...
   clientfd= arg;
@@ -559,6 +513,20 @@ void *stream(int arg, int groupindex, int index) {
       err_dump("can't open stream socket");
     }
 
+    /*
+    if -B is specified, balance tries to bind to it even on
+    outgoing connections 
+    */
+
+    if(outbindhost != NULL) {
+      struct sockaddr_in outbind_addr;
+      bzero((char *) &outbind_addr, sizeof(outbind_addr));
+      outbind_addr.sin_family      = AF_INET;
+      setipaddress(&outbind_addr.sin_addr, outbindhost);
+      if (bind(sockfd, (struct sockaddr *) &outbind_addr, sizeof(outbind_addr)) < 0) {
+      }
+    }
+
     b_readlock();
       bzero((char *) &serv_addr, sizeof(serv_addr));
       serv_addr.sin_family      = AF_INET;
@@ -566,7 +534,10 @@ void *stream(int arg, int groupindex, int index) {
       serv_addr.sin_port        = htons(chn_port(common,groupindex,index));
     b_unlock();
 
-    signal(SIGALRM, alarm_handler);
+    alrm_action.sa_handler = alrm_handler;
+    alrm_action.sa_flags = 0;			// don't restart !
+    sigemptyset(&alrm_action.sa_mask);
+    sigaction(SIGALRM,&alrm_action,NULL);
     alarm(connect_timeout);
   
     if (connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
@@ -588,21 +559,32 @@ void *stream(int arg, int groupindex, int index) {
       b_readlock();
         for(;;) {
 	  for(;;) {
-            index++;
-	    if(index >= grp_nchannels(common,groupindex)) {
-	      index=0;
-	    }
-	    if(index == startindex) {
-	      index = -1;	// Giveup 
-	      break;
-	    }
-            if(chn_status(common,groupindex,index) == 1 &&
-             (chn_maxc(common,groupindex,index) == 0 ||
-              (chn_c(common, groupindex, index) < chn_maxc(common, groupindex, index)))) {
-	      break;		// new index found 
-	    } else {
-	      continue;
-	    }
+            if(grp_type(common,groupindex) == GROUP_RR) {
+	      // HIER
+              // 
+              index++;
+	      if(index >= grp_nchannels(common,groupindex)) {
+	        index=0;
+	      }
+	      if(index == startindex) {
+	        index = -1;	// Giveup 
+	        break;
+	      }
+              if(chn_status(common,groupindex,index) == 1 &&
+               (chn_maxc(common,groupindex,index) == 0 ||
+                (chn_c(common, groupindex, index) < chn_maxc(common, groupindex, index)))) {
+	        break;		// new index found 
+	      } else {
+	        continue;
+	      }
+ 	    } else if (grp_type(common,groupindex) == GROUP_HASH) {
+
+              // If the current group is type hash, we giveup immediately 
+              index=-1;
+              break;
+            } else {
+              err_dump("PANIC: invalid group in stream()");
+            }
 	  }
 
           if(index >= 0) {
@@ -615,15 +597,38 @@ again:
               // giveup, index=-1.
               break;
             } else {
-              if(grp_nchannels(common,groupindex) > 0) {
-                index=grp_current(common,groupindex);
-                startindex=index; 	// This fixes the "endless loop error"
-					// with all hosts being down and one
-					// in the last group... (from Anthony Baxter) 
+              if(grp_type(common,groupindex) == GROUP_RR) {
+
+                if(grp_nchannels(common,groupindex) > 0) {
+                  index=grp_current(common,groupindex);
+                  startindex=index; 	// This fixes the "endless loop error"
+					  // with all hosts being down and one
+					  // in the last group... (from Anthony Baxter) 
+                } else {
+                  goto again;
+                }
+                break; 
+              } else if(grp_type(common,groupindex) == GROUP_HASH) {
+		unsigned int uindex;
+                uindex=hash_fold(client_address, client_address_size); 
+
+                if(debugflag) 
+   	          fprintf(stderr, "fold returns %d\n", uindex);
+	        index=uindex % grp_nchannels(common,groupindex);
+                if(debugflag) 
+   	          fprintf(stderr, "modulo %d gives %d\n",grp_nchannels(common,groupindex), index);
+
+                if(chn_status(common,groupindex,index) == 1 && 
+                   (chn_maxc(common,groupindex,index) == 0 ||
+                    (chn_c(common,groupindex,index) < chn_maxc(common,groupindex,index)))
+                   ) {
+                   break;
+                } else {
+                  goto again;	// next group !
+                }
               } else {
-                goto again;
+                err_dump("PANIC: invalid group in stream()");
               }
-              break; 
             }
           }
         }
@@ -643,18 +648,19 @@ again:
       }
 
     } else {
+      alarm(0);     // Cancel the alarm since we successfully connected
       if(debugflag) {
-	fprintf(stderr, "connect to channel %d succesful\n", index);
+	fprintf(stderr, "connect to channel %d successful\n", index);
       }
 
       // this prevents the 'channel 2 overload problem'
 
       b_writelock();
-      grp_current(common,groupindex)=index;
-      grp_current(common,groupindex)++;
-      if(grp_current(common,groupindex) >= grp_nchannels(common,groupindex)) {
-        grp_current(common,groupindex)=0;
-      }
+        grp_current(common,groupindex)=index;
+        grp_current(common,groupindex)++;
+        if(grp_current(common,groupindex) >= grp_nchannels(common,groupindex)) {
+          grp_current(common,groupindex)=0;
+        }
       b_unlock();
 
       // everything's fine ... 
@@ -668,43 +674,53 @@ again:
   exit(0);
 }
 
-
-static usage() {
+static void initialize_release_variables() {
   char* version;
   char* revision_copy;
   char* token;
 
   if((revision_copy=(char*) malloc(strlen(revision)+1))==NULL) {
-    fprintf(stderr,"malloc problem in usage()\n");
+    fprintf(stderr,"malloc problem in initialize_release_variables()\n");
   } else {
     strcpy(revision_copy, revision);
     token=strtok(revision_copy," ");
     token=strtok(NULL," ");
-    version=token!=NULL?token:"*undefined*";
-    fprintf(stderr,"\n");
-
-    fprintf(stderr,"balance %s - a balancing tcp proxy\n\n",version);
-
-    fprintf(stderr,"Copyright (c) 2000,2001 by Thomas Obermair (obermair@acm.org) \n"); 
-    fprintf(stderr,"and Inlab Software GmbH (info@inlab.de), Gruenwald, Germany.\n"); 
-    fprintf(stderr,"All rights reserved.\n");
-    fprintf(stderr,"\n");
+    version=token!=NULL?token:"0.0";
+    release=atoi(version);
+    if(strlen(version) >=3 ) {
+      subrelease=atoi(version+2);
+    } else {
+      subrelease=0;
+    }
     free(revision_copy);
   }
+}
 
-  fprintf(stderr,"usage:\n");
-  fprintf(stderr,"  balance [-b host] [-t sec] [-dfp] \\\n");
-  fprintf(stderr,"          port [h1[:p1[:maxc1]] [!] [ ... hN[:pN[:maxcN]]]]\n");
-  fprintf(stderr,"  balance [-b host] [-id] port\n");
+void static usage() {
+  fprintf(stderr,"\n");
+  fprintf(stderr,"balance %d.%d\n", release, subrelease);
+  fprintf(stderr,"Copyright (c) 2000-2003,2004 by Inlab Software GmbH, Gruenwald, Germany.\n"); 
+  fprintf(stderr,"All rights reserved.\n");
   fprintf(stderr,"\n");
 
-  fprintf(stderr,"  -b host   bind to specific host address\n");
+  fprintf(stderr,"usage:\n");
+  fprintf(stderr,"  balance [-b host] [-t sec] [-T sec] [-dfp] \\\n");
+  fprintf(stderr,"          port [h1[:p1[:maxc1]] [!] [ ... hN[:pN[:maxcN]]]]\n");
+  fprintf(stderr,"  balance [-b host] -i [-d] port\n");
+  fprintf(stderr,"  balance [-b host] -c cmd  [-d] port\n");
+  fprintf(stderr,"\n");
+
+  fprintf(stderr,"  -b host   bind to specific host address on listen\n");
+  fprintf(stderr,"  -B host   bind to specific host address for outgoing connections\n");
+  fprintf(stderr,"  -c cmd    execute specified interactive command\n");
   fprintf(stderr,"  -d        debugging on\n");
   fprintf(stderr,"  -f        stay in foregound\n");
   fprintf(stderr,"  -i        interactive control\n");
   fprintf(stderr,"  -p        packetdump\n");
   fprintf(stderr,"  -t sec    specify connect timeout in seconds (default=%d)\n",DEFAULTTIMEOUT);
-  fprintf(stderr,"   !        separates channelgroups\n");
+  fprintf(stderr,"  -T sec    timeout (seconds) for select (0 => never) (default=%d)\n",DEFAULTSELTIMEOUT);
+  fprintf(stderr,"   !        separates channelgroups (declaring previous to be Round Robin)\n");
+  fprintf(stderr,"   %%        as !, but declaring previous group to be a Hash Type\n");
 
   fprintf(stderr,"\n");
   fprintf(stderr,"example:\n");
@@ -717,7 +733,7 @@ static usage() {
 
 // detach from controlling tty ...
 
-hintergrund() {
+void hintergrund() {
   int childpid;
   if ((childpid = fork()) < 0) {
     fprintf(stderr,"cannot fork\n");
@@ -727,12 +743,11 @@ hintergrund() {
       exit(0);  /* parent */
     }
   }
-#ifdef MYBSD
+#ifdef BalanceBSD
   setpgid(getpid(),0);
 #else
   setpgrp();
 #endif
-  signal(SIGCHLD,chld_handler);
 }
 
 COMMON* makecommon(int argc, char** argv, int source_port) {
@@ -763,9 +778,14 @@ COMMON* makecommon(int argc, char** argv, int source_port) {
       exit(1);
     }
 
+    mycommon->pid=getpid();
+    mycommon->release=release;
+    mycommon->subrelease=subrelease;
+
     for(group=0; group<MAXGROUPS; group++) {
       grp_nchannels(mycommon,group)=0;
       grp_current(mycommon,group)=0;
+      grp_type(mycommon,group)=GROUP_RR; 				// Default: RR
     }
 
     group=0;
@@ -773,6 +793,19 @@ COMMON* makecommon(int argc, char** argv, int source_port) {
 
     for(i=1; i<argc; i++) {
       if(!strcmp(argv[i],"!")) {
+        // This is a normal "GROUP_RR"-Type of Group
+        grp_type(mycommon,group) = GROUP_RR;
+        group++;
+        channel=0;
+        if(group>=MAXGROUPS) {
+          err_dump("groups exceed MAXGROUPS");
+        }
+        if(channel>=MAXCHANNELS) {
+          err_dump("to many channels in one group");
+        }
+      } else if(!strcmp(argv[i],"%")) {
+        // This is a "GROUP_HASH"
+        grp_type(mycommon,group) = GROUP_HASH;
         group++;
         channel=0;
         if(group>=MAXGROUPS) {
@@ -816,7 +849,7 @@ COMMON* makecommon(int argc, char** argv, int source_port) {
   return(mycommon);
 }
 
-mycmp(char* s1, char* s2) {
+int mycmp(char* s1, char* s2) {
   int l;
   l=strlen(s1)<strlen(s2)?strlen(s1):strlen(s2);
   if(strlen(s1) > strlen(s2)) {
@@ -826,18 +859,43 @@ mycmp(char* s1, char* s2) {
   }
 }
 
-shell() {
+int shell(char* argument) {
   int i;
   int currentgroup=0;
   char line[MAXINPUTLINE];
   char *command;
 
+  if(common->release == 0) {
+    printf("no master process, exiting.\n");
+    exit(2);
+  }
+
+  if(common->release != release || common->subrelease != subrelease) {
+    printf("release mismatch, expecting %d.%d, got %d.%d, exiting.\n",
+		    release,subrelease,common->release,common->subrelease);
+    exit(2);    
+  } 
+
+  if(kill(common->pid, SIGUSR1) == -1) {
+    printf("no master process with pid %d, exiting.\n",common->pid);
+    exit(2);   
+  }	   
+
+  if(argument == NULL) {
+    printf("\nbalance %d.%d interactive command shell\n",release, subrelease);
+    printf("PID of master process is %d\n\n", common->pid);
+  }     
+
   for(;;) {
     
-    printf("balance[%d] ",currentgroup);
-    if (fgets(line, MAXINPUTLINE, stdin) == NULL) {
-      printf("\n");
-      exit(0);
+    if(argument == NULL) { 
+      printf("balance[%d] ",currentgroup);
+      if (fgets(line, MAXINPUTLINE, stdin) == NULL) {
+        printf("\n");
+        exit(0);
+      }
+    } else {
+      strncpy(line, argument, MAXINPUTLINE);
     }
 
     if((command=strtok(line," \t\n")) != NULL) {
@@ -848,12 +906,13 @@ shell() {
           {
             int group;
 
-	    printf("%3s %2s %3s %16s %5s %4s %4s %11s %11s\n",
-		   "GRP", "#", "S","ip-address","port","c","maxc","sent","rcvd"); 
+	    printf("%3s %4s %2s %3s %16s %5s %4s %4s %11s %11s\n",
+		   "GRP", "Type", "#", "S","ip-address","port","c","maxc","sent","rcvd"); 
             for (group=0; group<=MAXGROUPS; group++) {
 	      for(i=0; i<grp_nchannels(common,group); i++) {
-	        printf("%3d %2d %3s %16s %5d %4d %4d %11u %11u\n", 
+	        printf("%3d %4s %2d %3s %16s %5d %4d %4d %11u %11u\n", 
                        group,
+                       grp_type(common,group)==GROUP_RR?"RR":"Hash",
 		       i, 
 		       chn_status(common,group,i)==1?"ENA":"dis",
 		       inet_ntoa(chn_ipaddr(common,group,i)), 
@@ -874,15 +933,29 @@ shell() {
 	printf("  disable <channel>     disables specified channel in current group\n");
 	printf("  enable <channel>      enables channel in current group\n");
 	printf("  group <group>         changes current group to <group>\n");
+	printf("  hash                  sets distribution scheme of current group to Hash\n");
 	printf("  help                  prints this message\n");
+	printf("  kill                  kills master process and quits interactive mode\n");
 	printf("  maxc <channel> <maxc> specifies new maxc for channel of current group\n");
 	printf("  quit                  quit interactive mode\n");
 	printf("  reset <channel>       reset all counters of channel in current group\n");
+	printf("  rr                    sets distribution scheme of current group to Round Robin\n");
 	printf("  show                  show all channels in all groups\n");
 	printf("  version               show version id\n");
 
+      } else if(mycmp(command,"kill")) {
+        kill(common->pid, SIGKILL);
+        sleep(1);
+        if(kill(common->pid, SIGUSR1) == -1) {
+	  printf("shutdown complete, exiting.\n");
+	  common->release=0;
+	  exit(2); 
+        } else {
+	  printf("shutdown failed.\n");
+        }		
       } else if(mycmp(command,"disable")) {
-        char* arg, n; 
+        char* arg; 
+	int n;
         if((arg=strtok(NULL," \t\n")) != NULL) {
 	  n=atoi(arg);
 	  if(n<0 || n >= grp_nchannels(common,currentgroup)) {
@@ -893,6 +966,7 @@ shell() {
 		printf("channel %d already disabled\n",n);
 	      } else {
 		chn_status(common,currentgroup,n) = 0;
+		printf("channel %d disabled\n",n);
 	      }
 	    c_unlock(0, n);
 	  }
@@ -913,7 +987,8 @@ shell() {
 	}
 
       } else if(mycmp(command,"reset")) {	// reset channel counters
-        char* arg, n; 
+        char* arg; 
+	int n;
 
         if((arg=strtok(NULL," \t\n")) != NULL) {
 	  n=atoi(arg);
@@ -924,6 +999,7 @@ shell() {
 		chn_breceived(common, currentgroup, n)=0;
 		chn_bsent(common, currentgroup, n)=0;
 	    c_unlock(currentgroup, n);
+	    printf("channel %d counters resetted\n", n);
 	  }
 	} else {
 	  printf("syntax error\n");
@@ -931,7 +1007,8 @@ shell() {
 
       } else if(mycmp(command,"enable")) {
 
-        char* arg, n; 
+        char* arg; 
+	int n;
         if((arg=strtok(NULL," \t\n")) != NULL) {
 	  n=atoi(arg);
 	  if(n<0 || n >= grp_nchannels(common,currentgroup)) {
@@ -942,6 +1019,7 @@ shell() {
 		printf("channel %d already enabled\n",n);
 	      } else {
 		chn_status(common,currentgroup,n)=1;
+		printf("channel %d enabled\n",n);
 	      }
 	    c_unlock(currentgroup, n);
 	  }
@@ -965,6 +1043,7 @@ shell() {
                 chn_bsent(common,currentgroup,grp_nchannels(common,currentgroup))=0;
                 chn_breceived(common,currentgroup,grp_nchannels(common,currentgroup))=0;
                 grp_nchannels(common,currentgroup)++;
+		printf("channel created\n");
 	      } else {
 		printf("invalid address\n");
 	      }
@@ -986,7 +1065,8 @@ shell() {
             if(atoi(arg1) < 0 || atoi(arg1) >= MAXCHANNELS || atoi(arg1)+1 > grp_nchannels(common,currentgroup)) { 
 	      printf("unknown channel\n");
             } else {
-	      chn_maxc(common,currentgroup,atoi(arg1))=atoi(arg2);;
+	      chn_maxc(common,currentgroup,atoi(arg1))=atoi(arg2);
+	      printf("maxc of channel %d changed to %d\n",atoi(arg1),atoi(arg2));
             } 
           } else {
 	    printf("syntax error\n");
@@ -998,60 +1078,78 @@ shell() {
         b_unlock();
 
       } else if(mycmp(command,"version")) {
-	printf("  $Id: balance.c,v 2.33 2001/09/15 22:00:54 tommy Exp $\n");
+	printf("  This is balance %d.%d\n", release, subrelease);
         printf("  MAXGROUPS=%d\n", MAXGROUPS);
         printf("  MAXCHANNELS=%d\n", MAXCHANNELS);
+      } else if(mycmp(command,"hash")) {
+        b_writelock();
+        grp_type(common,currentgroup) = GROUP_HASH;
+        b_unlock();
+	printf("group %d set to hash\n", currentgroup);
+
+      } else if(mycmp(command,"rr")) {
+        b_writelock();
+        grp_type(common,currentgroup) = GROUP_RR;
+        b_unlock();
+	printf("group %d set to round robin\n", currentgroup);
+
       } else {
         printf("syntax error\n");
       }
+      printf("\n");
     }
+    if (argument != NULL)
+      exit(0);
   }
 }
 
-#ifdef MYBSD
-void sig_chld(int signo) 
-{
-    int status, child_val;
-    if (waitpid(-1, &status, WNOHANG) < 0) 
-    {
-        fprintf(stderr, "waitpid failed\n");
-        return;
-    }
-}
-#endif
-
 char bindhost_address[FILENAMELEN];
 
-main(argc, argv)
-int	argc;
-char	*argv[];
+int main(int argc, char *argv[])
 {
   int			sockfd, newsockfd, clilen, childpid;
+  int sockopton = 1;
+  int			c;
+  int			source_port;
+  int fd;
+  char*			argument=NULL;
+  struct stat buffer;
   struct sockaddr_in	cli_addr, serv_addr;
-  char*			bindhost=NULL;
-
-#ifdef MYBSD
-  struct sigaction act;
+  struct sigaction usr1_action,chld_action;
+#ifdef BalanceBSD
   pid_t pid;
 #else
   struct rlimit 	r;
 #endif
-  int			c;
-  int			source_port;
 
   connect_timeout=DEFAULTTIMEOUT;
+  initialize_release_variables();
 
-  while((c = getopt(argc,argv,"b:t:dfpi")) != EOF) {
+  while((c = getopt(argc,argv,"c:b:B:t:T:dfpi")) != EOF) {
     switch(c) {
     case 'b':
       bindhost=optarg;
       break;
+    case 'B':
+      outbindhost=optarg;
+      break;
+    case 'c':
+      argument=optarg;
+      interactive=1;
+      foreground=1;
+      packetdump=0;
+      break;
     case 't':
       connect_timeout=atoi(optarg);
       if(connect_timeout < 1) {
-	usage;
+	usage();
       }
       break;
+    case 'T':
+      sel_tmout.tv_sec=atoi(optarg);
+      if(sel_tmout.tv_sec < 1)
+        usage();
+      break;
     case 'f':
       foreground=1;
       break;
@@ -1094,10 +1192,18 @@ char	*argv[];
       usage();
     }
   }
+  usr1_action.sa_handler = usr1_handler;
+  usr1_action.sa_flags= SA_RESTART;
+  sigemptyset(&usr1_action.sa_mask);
+  sigaction(SIGUSR1,&usr1_action,NULL);
 
+  chld_action.sa_handler = chld_handler;
+  chld_action.sa_flags= SA_RESTART;
+  sigemptyset(&chld_action.sa_mask);
+  sigaction(SIGCHLD,&chld_action,NULL);
   // really dump core if something fails... 
 
-#ifdef MYBSD
+#ifdef BalanceBSD
 #else
   getrlimit(RLIMIT_CORE, &r);
   r.rlim_cur=r.rlim_max;
@@ -1129,40 +1235,35 @@ char	*argv[];
   }
   serv_addr.sin_port=htons(source_port);
 
-  // evtl. Anlage des dummy-Files fuer ftok() und fctl()
-
-  {
-     struct stat buffer;
-     int fd;
-
-     sprintf(rendezvousfile,"%sbalance.%d.%s",SHMDIR,source_port,bindhost_address);
-
-     if(stat(rendezvousfile, &buffer) == -1) {
-       // File existiert (noch) nicht ...
-       if((fd = open(rendezvousfile, O_CREAT | O_RDWR, 0666)) == -1) {
-	 fprintf(stderr,"cannot create rendevouz file %s\n", rendezvousfile);
-	 exit(1);
-       } else {
-	 if(debugflag) 
-	   fprintf(stderr,"file %s created\n",rendezvousfile);
-	 close(fd);
-       }
-     } else {
-       if(debugflag) 
-	 fprintf(stderr,"file %s already there\n",rendezvousfile);
-     }
+  stat(SHMDIR, &buffer);
+  if (!S_ISDIR(buffer.st_mode)) {
+    fprintf(stderr,"ERROR: balance rendezvous directory does not exist, please create %s writable by this user\n", SHMDIR);
+	exit(1);
   }
 
-  // --
+  sprintf(rendezvousfile,"%sbalance.%d.%s",SHMDIR,source_port,bindhost_address);
 
-  signal(SIGCHLD, chld_handler);
+  if(stat(rendezvousfile, &buffer) == -1) {
+    // File not existing yet ...
+    if((fd = open(rendezvousfile, O_CREAT | O_RDWR, 0666)) == -1) {
+      fprintf(stderr,"cannot create rendezvous file %s\n", rendezvousfile);
+      exit(1);
+    } else {
+      if(debugflag) 
+        fprintf(stderr,"file %s created\n",rendezvousfile);
+      close(fd);
+    }
+  } else {
+    if(debugflag) 
+      fprintf(stderr,"file %s already exists\n",rendezvousfile);
+  }
 
   if(!foreground) {
     hintergrund();
   }
 
   if(interactive) {
-    // Kommandomodus !
+    // command mode ! 
 
     if((rendezvousfd=open(rendezvousfile,O_RDWR,0)) < 0) {
       perror("open");
@@ -1172,7 +1273,8 @@ char	*argv[];
       fprintf(stderr,"cannot alloc COMMON struct\n");
       exit(1);
     }
-    shell();
+      
+    shell(argument);
   }
 
   // Open a TCP socket (an Internet stream socket).
@@ -1181,14 +1283,11 @@ char	*argv[];
     err_dump("can't open stream socket");
   }
 
-  { 
-    int sockopton = 1;
     if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, 
          (char *)&sockopton, sizeof(int) ) == -1 ) {
       perror("setsockopt(SO_REUSEADDR)");
       err_dump("can't so_reuseaddr");
     }
-  }
 
   if(debugflag) {
     fprintf(stderr,"bindhost_adress=[%s]\n",bindhost_address);
@@ -1204,27 +1303,11 @@ char	*argv[];
 
   listen(sockfd, SOMAXCONN);
 
-/*
-#ifdef MYBSD
-  act.sa_handler = sig_chld;
-  sigemptyset(&act.sa_mask);
-  act.sa_flags = SA_NOCLDSTOP|SA_NOCLDWAIT;
-  if (sigaction(SIGCHLD, &act, NULL) < 0) {
-    fprintf(stderr, "sigaction failed\n");
-    return 1;
-  }
-#else
-#endif
-*/
-
   for (;;) {
     int index;
+    unsigned int uindex;
     int groupindex=0;	// always start at groupindex 0
 
-    // ... richards original comment left intact:  
-    // Wait for a connection from a client process.
-    // This is an example of a concurrent server.
-
     clilen = sizeof(cli_addr);
 
     newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
@@ -1235,42 +1318,80 @@ char	*argv[];
       continue;
     }
 
-    // the balancing itself ...
+    if(debugflag) { 
+		  fprintf(stderr, "connect from %s\n", inet_ntoa(cli_addr.sin_addr)); 
+		}
 
-    b_writelock();
+		// the balancing itself ...
+		// groupindex = 0
+		// hier wird entschieden, mit welchem "channel" es erstmalig versucht wird !
+		// Client-Adresse verfuegbar in cli_addr.sin_addr
 
-      for(;;) {
+		b_writelock();
 
-        index = grp_current(common,groupindex);
-        for(;;) {
-          if(chn_status(common,groupindex,index) == 1 && 
-             (chn_maxc(common,groupindex,index) == 0 ||
-              (chn_c(common,groupindex,index) < chn_maxc(common,groupindex,index)))
-             ) {
-	    break; 						// channel found
-	  } else {
-	    index++;
-	    if(index >= grp_nchannels(common,groupindex)) {
-	      index=0;
-	    }
-	    if(index == grp_current(common,groupindex)) {
-	      index =-1;					// no channel available in this group
+		  for(;;) {
+
+			index = grp_current(common,groupindex);
+			for(;;) {
+			  if(grp_type(common,groupindex) == GROUP_RR) {
+				if(chn_status(common,groupindex,index) == 1 && 
+				   (chn_maxc(common,groupindex,index) == 0 ||
+					(chn_c(common,groupindex,index) < chn_maxc(common,groupindex,index)))
+				   ) {
+			  break; 						// channel found
+			} else {
+			  index++;
+			  if(index >= grp_nchannels(common,groupindex)) {
+				index=0;
+			  }
+			  if(index == grp_current(common,groupindex)) {
+				index =-1;					// no channel available in this group
+				break;
+			  }
+			}
+			  } else if (grp_type(common,groupindex) == GROUP_HASH) {
+				// ---------------------------------------------------------------------
+				// Hash !
+			// HIER
+            uindex=hash_fold((char*) &cli_addr.sin_addr, sizeof(cli_addr.sin_addr));
+            if(debugflag) 
+   	      fprintf(stderr, "fold returns %d\n", uindex);
+	    index=uindex % grp_nchannels(common,groupindex);
+            if(debugflag) 
+   	      fprintf(stderr, "modulo %d gives %d\n",grp_nchannels(common,groupindex), index);
+            if(chn_status(common,groupindex,index) == 1 && 
+               (chn_maxc(common,groupindex,index) == 0 ||
+                (chn_c(common,groupindex,index) < chn_maxc(common,groupindex,index)))
+               ) {
+               break;						// channel found, channel valid for HASH
+            } else {
+	      index =-1;
 	      break;
-	    }
-	  }
+            }	
+          } else {
+            err_dump("PANIC: invalid group type");
+          }
         }
+
+        // Hier fallen wir "raus" mit dem index in der momentanen Gruppe, oder -1
+        // wenn nicht moeglich in dieser Gruppe
+
         grp_current(common,groupindex)=index;
-        grp_current(common,groupindex)++;
+        grp_current(common,groupindex)++;	// current index dieser gruppe wieder null, wenn vorher ungueltig (-1)
+
+        // Der index der gruppe wird neu berechnet und gespeichert, "index" ist immer noch 
+        // -1 oder der zu waehlende index...
+
         if(grp_current(common,groupindex) >= grp_nchannels(common,groupindex)) {
           grp_current(common,groupindex)=0;
         }
 
         if(index >=0) {
-          chn_c(common,groupindex,index)++; 			// we promise a succesful connection 
+          chn_c(common,groupindex,index)++; 			// we promise a successful connection 
 								// c++ 
           break;						// index in this group found 
         } else {
-          groupindex++;
+          groupindex++;						// try next group !
           if(groupindex >= MAXGROUPS) {
             break;	// end of groups...
           }
@@ -1290,7 +1411,9 @@ char	*argv[];
 	}
       } else if (childpid == 0) {				// child process 
         close(sockfd);						// close original socket 
-        stream(newsockfd, groupindex, index);			// process the request 
+								// process the request: 
+
+        stream(newsockfd, groupindex, index, (char*) &cli_addr.sin_addr, sizeof(cli_addr.sin_addr)); 
         exit(0);
       }
     }
